/// A library for Dart developers that want to interact with the JMB Aqualight.
///
/// This library communicates over UDP with your controller. Make sure that your
/// are connected to the controller's WiFi network.
///
/// This library is currently used by the JMB Aqualight apps. Available for iOS
/// and Android.
library jmb_aqualight;

import 'src/packets-discovery.dart' show createDiscoveryPacket;
import 'src/packets-live.dart' show createLivePackets;
import 'src/packets-schedule.dart' show createSchedulePackets;
import 'src/packets-sync.dart' show createTimeSyncPacket;
import 'src/send.dart' show sendPackets;

/// The connection information.
class ConnectionInfo {
  /// The unique bytes that identify the controller. This needs to be a String
  /// with length 24.
  String controllerId;

  /// Default constructor.
  ConnectionInfo(this.controllerId) {
    if (controllerId.length != 24) {
      throw ('The controller\'s id needs to have a length of 24');
    }
  }
}

/// Connects with the JMB Aqualight controller.
///
/// Make sure that you are connected to the controller's WiFi network.
///
/// Throws an error if the controller cannot be reached.
Future<ConnectionInfo> connect() async {
  final packets = [createDiscoveryPacket()];
  final response = await sendPackets(packets);
  final controllerId = response[0].substring(8, 32);
  return ConnectionInfo(controllerId);
}

/// Synchronizes your system time with the controller.
///
/// Throws an error if the controller cannot be reached.
Future<void> timeSync(ConnectionInfo connectionInfo) async {
  final packets = [createTimeSyncPacket(connectionInfo.controllerId)];
  await sendPackets(packets);
}

/// Plays a live schedule on the controller (preview mode).
///
/// The schedule needs to be a list of 48 entries (one per half hour). Each
/// entry should contain 5 values between 0-100 (inclusive), one for each
/// channel.
///
/// The smooth parameter is used to add some smoothness to the preview. Only
/// used to be able to debug on the sending device.
///
/// Throws an error if the controller cannot be reached or when the schedule is
/// invalid.
Future<void> liveSchedule(
    ConnectionInfo connectionInfo, List<List<int>> schedule,
    [smooth = true]) async {
  _validateSchedule(schedule);
  final packets =
      createLivePackets(connectionInfo.controllerId, schedule, smooth);
  await sendPackets(packets);
}

/// Uploads a schedule to the controller.
///
/// The schedule needs to be a list of 48 entries (one per half hour). Each
/// entry should contain 5 values between 0-100 (inclusive), one for each
/// channel.
///
/// Throws an error if the controller cannot be reached or when the schedule is
/// invalid.
Future<void> uploadSchedule(
    ConnectionInfo connectionInfo, List<List<int>> schedule) async {
  _validateSchedule(schedule);
  final packets = createSchedulePackets(connectionInfo.controllerId, schedule);
  await sendPackets(packets);
}

void _validateSchedule(List<List<int>> schedule) {
  if (schedule.length != 48) {
    throw ('Expected a schedule with 48 entries.');
  }

  for (var entry in schedule) {
    if (entry.length != 5) {
      throw ('Expected 5 values per entry in the schedule.');
    }

    for (var val in entry) {
      if (val < 0 || val > 100) {
        throw ('A value in the schedule needs to be between 0 and 100');
      }
    }
  }
}
