import 'dart:async';
import 'dart:io';
import 'convert.dart';
import 'packets-general.dart';

const IP = '192.168.4.1';
const BROADCAST_IP = '255.255.255.255';
const PORT = 5000;

Future<List<String>> sendPackets(List<Packet> packets, [broadcast = false]) {
  final completer = Completer<List<String>>();

  RawDatagramSocket.bind(InternetAddress.anyIPv4, PORT)
      .then((RawDatagramSocket socket) {
    print('Client listening ${socket.address.address}:${socket.port}');

    final ip = broadcast ? BROADCAST_IP : IP;
    final result = <String>[];
    Timer timeout;

    void closeSocket() {
      socket.close();
      print('Client closed');
      print('==============');
    }

    void timeoutHandler() {
      closeSocket();
      completer.completeError('Unable to reach controller');
    }

    void sendNextPacket() {
      if (packets.isNotEmpty) {
        final packet = packets.removeAt(0);

        // Keep this to spot development errors
        if (packet.data.length != 128) {
          print('ERROR Invalid packet: ${packet.data}');
        }

        print('Send: ${packet.data} (${ip}:${PORT})');

        socket.send(decode(packet.data), InternetAddress(IP), PORT);

        if (!packet.hasResponse) {
          // Send the next packet in 150 ms
          Timer(Duration(milliseconds: 150), sendNextPacket);
        } else {
          // We expect a response within 10 s
          timeout = Timer(Duration(seconds: 10), timeoutHandler);
        }
      } else {
        closeSocket();
        completer.complete(result);
      }
    }

    socket.listen((RawSocketEvent e) {
      final d = socket.receive();

      if (d == null) return;

      // In broadcast mode, we also receive our own broadcasted packets. Discard these.
      if (d.address.address != IP) return;

      final message = encode(d.data);
      print('Recv: ${message} (${d.address.address}:${d.port})');

      result.add(message);

      // We got our response, clear the timeout
      if (timeout != null) timeout.cancel();

      // Wait 300ms before sending the next packet.
      Timer(Duration(milliseconds: 300), sendNextPacket);
    });

    // Send the next packet in 150 ms
    Timer(Duration(milliseconds: 150), sendNextPacket);
  }).catchError((error) {
    completer.completeError(
        'Unable to bind to port ${PORT}. Please check that this port is available.');
  });

  return completer.future;
}
